"""
Demoes how to verify a generic tx, no a priori on its type.
"""

import json
import pytest
import sys

sys.path.append("../")
from polysign.signerfactory import SignerFactory


def check_generic_sig(name: str, file: str, verbose: bool=False):
    with open(file) as f:
        tx = json.load(f)
        # Caller is responsible for creating the bin buffer to check the sig against
        buffer = str(
            (
                tx["timestamp"],
                tx["address"],
                tx["recipient"],
                tx["amount"],
                tx["operation"],
                tx["openfield"],
            )
        ).encode("utf-8")
        if verbose:
            print(f"{name} buffer", buffer)

        SignerFactory.verify_bis_signature(
            tx["signature"], tx["public_key"], buffer, tx["address"]
        )
        if verbose:
            print(f"{name}: No Error")


def test_verify_rsa_signature1(verbose: bool=False):
    # Sample real rsa tx
    check_generic_sig("RSA 1", "./vectors/rsa_tx1.json", verbose)


def test_verify_rsa_signature2(verbose: bool=False):
    # Sample real rsa tx
    check_generic_sig("RSA 2", "./vectors/rsa_tx2.json", verbose)


def test_verify_rsa_fail_signature(verbose: bool=False):
    # Sample real rsa tx, with one byte changed, must raise
    with pytest.raises(ValueError) as execinfo:
        check_generic_sig("RSA", "./vectors/rsa_tx_fail.json", verbose)
        # The following may be too strong or need formal specs to make sure the message is right.
        assert str(execinfo.value) == "Bad Signature"


def test_verify_ecdsa_signature2(verbose: bool=False):
    # Sample fake rsa tx
    check_generic_sig("ECDSA 2", "./vectors/ecdsa_tx2.json", verbose)


def test_verify_ecdsa_fail_signature(verbose: bool=False):
    # Sample rsa tx, with one byte changed, must raise
    with pytest.raises(ValueError) as execinfo:
        check_generic_sig("ECDSA", "./vectors/ecdsa_tx_fail.json", verbose)
        assert str(execinfo.value) == "Bad Signature"


def test_verify_ed25519_signature2(verbose: bool=False):
    # Sample fake ed25519 tx
    check_generic_sig("ED25519 2", "./vectors/ed25519_tx2.json", verbose)


def test_verify_ed25519_fail_signature(verbose: bool=False):
    # Sample real ed25519 tx, with one byte changed, must raise
    with pytest.raises(ValueError) as execinfo:
        check_generic_sig("ed25519", "./vectors/ed25519_tx_fail.json", verbose)
        assert str(execinfo.value) == "Bad Signature"


if __name__ == "__main__":
    test_verify_rsa_signature1(True)
    test_verify_rsa_fail_signature(True)
    test_verify_ed25519_fail_signature(True)



"""
Test 100 random sigs of every scheme
"""

import json
import base64
import sys
import random

sys.path.append("../")
from polysign.signerfactory import SignerFactory
from polysign.signer import SignerType

SEED = "This has to be reproducible"


def generic_test(signer_type: SignerType.ECDSA, verbose: bool=False):
    if verbose:
        print("Type", signer_type)
    random.seed(SEED)
    for i in range(100):
        seed = hex(random.getrandbits(32 * 8))[2:]
        if verbose:
            # We can have leading zeros that get discarded. In that case, the seed will be shorter
            # and the lig will use it as a random seed to regenerate a full 32 bytes seed.
            # left as is, not filling here.
            print("seed", seed, len(seed))
        signer = SignerFactory.from_seed(seed, signer_type)
        signer_dict = signer.to_dict()
        if verbose:
            print(signer_dict)
        public_key_hex = signer.to_dict()["public_key"]
        public_key_net = base64.b64encode(bytes.fromhex(public_key_hex)).decode("utf-8")

        buffer = hex(random.getrandbits(64 * 8)).encode()
        signature = signer.sign_buffer_for_bis(buffer)
        SignerFactory.verify_bis_signature(
            signature, public_key_net, buffer, signer_dict["address"]
        )


def test_100_rsa(verbose: bool=False):
    # This is a complex issue. We could do it, but this is not safe enough for crypto.
    # do it for tests? other tests vectors should be enough for RSA.
    if verbose:
        print("RSA impl by pycryptodome does not support seeded key gen.")
    pass


def test_100_ecdsa(verbose: bool=False):
    generic_test(SignerType.ECDSA, verbose=verbose)


def test_100_ed25519(verbose: bool=False):
    generic_test(SignerType.ED25519, verbose=verbose)


if __name__ == "__main__":
    test_100_ecdsa(True)
    test_100_ed25519(True)



"""
Creates Bis and test addresses from BIP32 test vectors.
"""

import base58
import json
import sys

sys.path.append("../")
from polysign.signer import SignerType, SignerSubType
from polysign.signerfactory import SignerFactory

# Never to be used for real addresses - from https://en.bitcoin.it/wiki/BIP_0032_TestVectors

if __name__ == "__main__":
    # Test vector 1
    print("Test vector 1\n")
    with open("ecdsa_test1.json") as f:
        private_keys = json.load(f)
    for private_key in private_keys:
        print("Private Key  \n{}  ".format(private_key))
        signer = SignerFactory.from_seed(private_key, SignerType.BTC)
        print("BTC Address  \n{}  ".format(signer.to_dict()["address"]))
        signer = SignerFactory.from_seed(private_key, SignerType.ECDSA)
        print("BIS Address  \n{}  ".format(signer.to_dict()["address"]))
    # Test vector 2
    print("\nTest vector 2\n")
    with open("ecdsa_test2.json") as f:
        private_keys = json.load(f)
    for private_key in private_keys:
        print("Private Key  \n{}  ".format(private_key))
        signer = SignerFactory.from_seed(private_key, SignerType.BTC)
        print("BTC Address  \n{}  ".format(signer.to_dict()["address"]))
        signer = SignerFactory.from_seed(private_key, SignerType.ECDSA)
        print("BIS Address  \n{}  ".format(signer.to_dict()["address"]))



import re
from os import urandom
from typing import Type, Union

from polysign.signer import Signer, SignerType, SignerSubType
from polysign.signer_btc import SignerBTC
from polysign.signer_crw import SignerCRW
from polysign.signer_rsa import SignerRSA
from polysign.signer_ecdsa import SignerECDSA
from polysign.signer_ed25519 import SignerED25519

RE_RSA_ADDRESS = re.compile(r"^[abcdef0123456789]{56}$")
# Improve ECDSA/ED25519 by Bizzzy
RE_ECDSA_ADDRESS = re.compile(r"^Bis1[123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz]{28,52}$")


def signer_for_type(signer_type: SignerType) -> Union[Type[Signer], None]:
    """Returns the class matching a signer type."""
    links = {SignerType.RSA: SignerRSA, SignerType.ED25519: SignerED25519,
             SignerType.ECDSA: SignerECDSA, SignerType.BTC: SignerBTC,
             SignerType.CRW: SignerCRW,
             }
    return links.get(signer_type, None)


class SignerFactory:
    """"""

    @classmethod
    def from_private_key(cls, private_key: Union[bytes, str], signer_type: SignerType=SignerType.RSA,
                         subtype: SignerSubType=SignerSubType.MAINNET_REGULAR) -> Signer:
        """Detect the type of the key, creates and return the matching signer"""
        # TODO: detect by private_key
        signer_class = signer_for_type(signer_type)
        if signer_class is None:
            raise ValueError("Unsupported Key type")
        signer = signer_class()
        signer.from_private_key(private_key, subtype=subtype)
        return signer

    @classmethod
    def from_full_info(cls, private_key: Union[bytes, str], public_key: Union[bytes, str]=b'', address: str='',
                       signer_type: SignerType=SignerType.RSA, subtype: SignerSubType=SignerSubType.MAINNET_REGULAR,
                       verify: bool=True) -> Signer:
        pass

    @classmethod
    def address_to_signer(cls, address: str) -> Type[Signer]:
        if RE_RSA_ADDRESS.match(address):
            return SignerRSA
        elif RE_ECDSA_ADDRESS.match(address):
            if len(address) > 50:
                return SignerED25519
            else:
                return SignerECDSA

        raise ValueError("Unsupported Address type")

    @classmethod
    def address_is_valid(cls, address: str) -> bool:
        if RE_RSA_ADDRESS.match(address):
            # RSA, 56 hex
            return True
        elif RE_ECDSA_ADDRESS.match(address):
            if 50 < len(address) < 60:
                # ED25519, around 54
                return True
            if 30 < len(address) < 50:
                # ecdsa, around 37
                return True
        return False

    @classmethod
    def address_is_rsa(cls, address: str) -> bool:
        """Returns whether the given address is a legacy RSA one"""
        return RE_RSA_ADDRESS.match(address) is not None

    @classmethod
    def from_seed(cls, seed: str='', signer_type: SignerType=SignerType.RSA,
                  subtype: SignerSubType=SignerSubType.MAINNET_REGULAR) -> Signer:
        if seed == '':
            seed = urandom(32).hex()
        signer_class = signer_for_type(signer_type)
        if signer_class is None:
            raise ValueError("Unsupported Key type")
        signer = signer_class()
        signer.from_seed(seed, subtype)
        return signer

    @classmethod
    def verify_bis_signature(cls, signature: str, public_key: str, buffer: bytes, address: str, ) -> None:
        """Verify signature from bismuth tx network format"""
        # Find the right signer class
        verifier = cls.address_to_signer(address)
        # let it do the job
        verifier.verify_bis_signature(signature, public_key, buffer, address)

    @classmethod
    def verify_bis_signature_raw(cls, signature: bytes, public_key: bytes, buffer: bytes, address: str, ) -> None:
        """Verify signature from decoded - bin - sig and pubkey"""
        verifier = cls.address_to_signer(address)
        verifier.verify_bis_signature_raw(signature, public_key, buffer, address)



import base58
import base64
from time import time

NB = 100000

SIGNATURE_58 = "381yXZ8CZPvDn3WWyfCWYxqpGnbGQVxfRcUdpu6LbjycoTpYckPEoc6aorS5RiEZPh5Z6sm2BDLZPpHQ5FNhmH1SDYDiSZ9j"
SIGNATURE_HEX = base58.b58decode(SIGNATURE_58).hex()
SIGNATURE_64 = base64.b64encode(base58.b58decode(SIGNATURE_58)).decode("utf-8")

print(SIGNATURE_58, SIGNATURE_HEX, SIGNATURE_64)


def decode_58():
    for i in range(NB):
        dec = base58.b58decode(SIGNATURE_58)
    return dec


def decode_64():
    for i in range(NB):
        dec = base64.b64decode(SIGNATURE_64)
    return dec


def decode_hex():
    for i in range(NB):
        dec = bytes.fromhex(SIGNATURE_HEX)
    return dec


def timer(func):
    start = time()
    func()
    print(func, time() - start)


if __name__ == "__main__":
    timer(decode_58)
    timer(decode_64)
    timer(decode_hex)



#!/usr/bin/env python3

from math import log, ceil
import sys


# lookup tables to convert integers in the range [0, 58) to base-58 digits and back
int_to_b58_digit = "123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz"
b58_digit_to_int = {b58: i for i, b58 in enumerate(int_to_b58_digit)}


# convert a (long) integer to its base-58 representation string
def int_to_base58_str(int_rep):
    base58_str = ""
    while int_rep:
        int_rep, remainder = divmod(int_rep, 58)
        base58_str = int_to_b58_digit[remainder] + base58_str
    return base58_str


def prepended_bytes(prepended_b58_digits, b256_digit_count):

    # ones are a special case in base58check format;
    # count and remove them, they are added back in later
    ones = 0
    for b58 in prepended_b58_digits:
        if b58 != "1":
            break
        ones += 1
        prepended_b58_digits = prepended_b58_digits[1:]
    if not prepended_b58_digits:  # if they're all 1's
        return ones * b"\0"

    # calc the # of base58 digits required for b256_digit_count bytes of "real" data
    # (not including the prepended base58 digits)
    b58_digit_count = ceil(b256_digit_count * log(256) / log(58))

    do_overflow_check = True
    while True:
        # calc the minimum integral value which starts with the desired digits in base-58
        min_int = 0
        for b58 in prepended_b58_digits:
            min_int *= 58
            min_int += b58_digit_to_int[b58]
        #
        # left-shift (mult. by a power of 58) to be just left of the "real" base-58 data
        min_int *= 58 ** b58_digit_count

        # uncomment to sanity-check that min_int is correct
        # print(" min_int:", ones * '1' + int_to_base58_str(min_int))

        # right-shift by multiples of 8 bits (base-256) to retrieve only the
        # most-significant bytes which are to the left of the "real" base-256 data
        min_int >>= b256_digit_count * 8
        # right-shifing effectively rounds min_int down, but we
        # need it rounded up, so add one to round it up instead
        min_int += 1

        # because min_int has been rounded up above, it's possible that adding it to "real"
        # data could cause an overflow in base-58 making the prepended_b58_digits increment
        # by one; if this could happen, left-shift prepended_b58_digits and repeat
        if do_overflow_check:
            max_real_data_int = (1 << b256_digit_count * 8) - 1
            max_base58_str = int_to_base58_str(
                (min_int << b256_digit_count * 8) + max_real_data_int
            )
            if not max_base58_str.startswith(prepended_b58_digits):
                prepended_b58_digits += "1"
                do_overflow_check = (
                    False
                )  # it doesn't matter if the '1' appended above overflows to '2'

                # uncomment to confirm that the max possible value
                # wouldn't have the desired prepended base-58 digits
                # print("overflow:", ones * '1' + max_base58_str)

                continue

        # prepend any ones according to base58check rules, and convert min_int to a byte string
        return ones * b"\0" + min_int.to_bytes(
            length=(min_int.bit_length() + 7) // 8, byteorder="big"
        )


if __name__ == "__main__":
    if len(sys.argv) != 3:
        sys.exit(
            "usage: {} <STRING-TO-PREPEND> <DATA-BYTE-LEN (excluding 4-byte checksum)>".format(
                sys.argv[0]
            )
        )

    result = prepended_bytes(
        sys.argv[1], int(sys.argv[2]) + 4
    )  # add 4 for the checksum
    print(", ".join("{:#04x}".format(i) for i in result))

    """
    ECDSA (len 20 bytes + 4)
    Regular add
    Bis1    0x4f, 0x54, 0x5b
    Multisig/script
    Bism    0x4f, 0x54, 0xc8
    
    tesnet
    tBis    0x01, 0x7a, 0xb6, 0x85
    testnet multisig
    mBis    0x01, 0x46, 0xeb, 0xa5
    
    
    ED25519 (len 32 bytes + 4)
    Regular add
    Bis1    0x03, 0xb8, 0x6c, 0xf3 //  0x02, 0x07, 0x1a, 0xb5
    Multisig/script
    Bism    0x03, 0xb8, 0x72, 0x14
    
    tesnet
    tBis    0x11, 0xc2, 0xce, 0x7c
    testnet multisig
    mBis    0x0f, 0x54, 0xfd, 0x2d
        
    """



"""
Abstract and Factory class to handle various Bismuth signature and addresses schemes
"""

import json

from abc import ABC, abstractmethod
from enum import Enum
from typing import Union


class SignerType(Enum):
    """
    Possible signing schemes
    """
    NONE = 0
    RSA = 1
    ECDSA = 2
    ED25519 = 3
    BTC = 1000  # Tests
    CRW = 1001  # Tests


class SignerSubType(Enum):
    """
    Possible addresses subtype
    """
    NONE = 0
    MAINNET_REGULAR = 1
    MAINNET_MULTISIG = 2
    TESTNET_REGULAR = 3
    TESTNET_MULTISIG = 4


class Signer(ABC):

    # Slots allow to spare ram when there can be several instances
    __slot__ = ('_private_key', '_public_key', '_address', '_type', '_subtype', '_compressed', 'verbose')

    _address_versions = {SignerSubType.MAINNET_REGULAR: b'\x00'}

    def __init__(self, private_key: Union[bytes, str]=b'', public_key: Union[bytes, str]=b'', address: str='',
                 compressed: bool=True, subtype: SignerSubType=SignerSubType.MAINNET_REGULAR):
        self._private_key = private_key
        self._public_key = public_key
        self._address = address
        self._type = SignerType.NONE
        self._subtype = subtype
        self.verbose = False
        self._compressed = compressed

    @property
    def compressed(self):
        return self._compressed

    @property
    def type(self):
        """Name of the signer instance"""
        return self._type.name

    @abstractmethod
    def from_private_key(self, private_key: Union[bytes, str], subtype: SignerSubType=SignerSubType.MAINNET_REGULAR):
        pass

    @abstractmethod
    def from_full_info(self, private_key: Union[bytes, str], public_key: Union[bytes, str]=b'', address: str='',
                       subtype: SignerSubType = SignerSubType.MAINNET_REGULAR, verify: bool=True):
        pass

    @abstractmethod
    def from_seed(self, seed: str='', subtype: SignerSubType=SignerSubType.MAINNET_REGULAR):
        """Use seed == '' to generate a random key"""
        pass

    @classmethod
    @abstractmethod
    def public_key_to_address(cls, public_key: Union[bytes, str],
                              subtype: SignerSubType=SignerSubType.MAINNET_REGULAR) -> str:
        """Reconstruct an address from the public key"""
        pass

    @classmethod
    def address_version_for_subtype(cls, subtype: SignerSubType) -> bytes:
        # Specific one if exists, else mainnet regular, else \x00
        return cls._address_versions.get(subtype, cls._address_versions.get(subtype.MAINNET_REGULAR, b'\x00'))

    @classmethod
    @abstractmethod
    def verify_signature(cls, signature: Union[bytes, str], public_key: Union[bytes, str], buffer: bytes,
                         address: str=''):
        """Verify signature from raw signature & pubkey. Address may be used to determine the sig type"""
        pass

    @classmethod
    @abstractmethod
    def verify_bis_signature(cls, signature: str, public_key: str, buffer: bytes, address: str='') -> None:
        """Verify signature from bismuth tx network format
        pubkey is b64 encoded twice - ecdsa and ed25519 are b64 encoded)"""
        pass

    @classmethod
    def verify_bis_signature_raw(cls, signature: bytes, public_key: bytes, buffer: bytes, address: str='') -> None:
        """Verify signature from bismuth tx network format
        pubkey is b64 encoded twice - ecdsa and ed25519 are b64 encoded)"""
        print("Abstract Method verify_bis_signature_raw")

    @abstractmethod
    def sign_buffer_raw(self, buffer: bytes) -> bytes:
        """Sign a buffer, sends a raw bytes array"""
        pass

    @abstractmethod
    def sign_buffer_for_bis(self, buffer: bytes) -> str:
        """Sign a buffer, sends under the format expected by bismuth network format"""
        pass

    def to_dict(self):
        """Returns core properties as dict, compact bin form"""
        info = {'address': self._address, 'private_key': self._private_key, 'public_key': self._public_key,
                'compressed': self._compressed, 'type': self._type.name, 'sub_type': self._subtype.name}
        return info

    def to_json(self):
        """Returns a json string, with bin items as hex strings"""
        info = self.to_dict()
        info['private_key'] = info['private_key'].hex()
        info['public_key'] = info['public_key'].hex()
        return json.dumps(info)



"""

"""

import hashlib
import random
from base64 import b64decode, b64encode
from hashlib import sha256
from os import urandom
from typing import Union

import base58
from coincurve import PrivateKey, verify_signature
from polysign.signer import Signer, SignerType, SignerSubType


class SignerECDSA(Signer):

    __slots__ = ('_key', )

    _address_versions = {SignerSubType.MAINNET_REGULAR: b'\x4f\x54\x5b',
                         SignerSubType.MAINNET_MULTISIG: b'\x4f\x54\xc8',
                         SignerSubType.TESTNET_REGULAR: b'\x01\x7a\xb6\x85',
                         SignerSubType.TESTNET_MULTISIG: b'\x01\x46\xeb\xa5'}

    def __init__(self, private_key: Union[bytes, str]=b'', public_key: Union[bytes, str]=b'', address: str='',
                 compressed: bool=True, subtype: SignerSubType=SignerSubType.MAINNET_REGULAR):
        super().__init__(private_key, public_key, address, compressed=compressed, subtype=subtype)
        self._key = None
        self._type = SignerType.ECDSA

    def from_private_key(self, private_key: Union[bytes, str], subtype: SignerSubType=SignerSubType.MAINNET_REGULAR):
        """Accepts both bytes[32] or str (hex format)"""
        if subtype != SignerSubType.MAINNET_REGULAR:
            self._subtype = subtype
        if type(private_key) == str:
            return self.from_seed(private_key, subtype=self._subtype)
        return self.from_seed(private_key.hex())

    def from_full_info(self, private_key: Union[bytes, str], public_key: Union[bytes, str]=b'', address: str='',
                       subtype: SignerSubType = SignerSubType.MAINNET_REGULAR, verify: bool=True):
        raise ValueError("SignerRsa.from_full_info not impl.")

    def from_seed(self, seed: str='', subtype: SignerSubType=SignerSubType.MAINNET_REGULAR):
        """Creates key from seed - for ecdsa, seed = pk - 32 bytes random buffer"""
        if subtype != SignerSubType.MAINNET_REGULAR:
            self._subtype = subtype
        if len(seed) > 64:
            # Too long seed, trim (could use better scheme for more entropy)
            seed = seed[:64]
        elif seed == '':
            # No seed, use urandom
            seed = urandom(32)
        elif len(seed) < 64:
            # Too short seed, use as PRNG seed
            random.seed(seed)
            seed = hex(random.getrandbits(32*8))[2:]
            while len(seed) < 64:
                seed = '0' + seed
            assert len(seed) == 64
        try:
            key = PrivateKey.from_hex(seed)
            public_key = key.public_key.format(compressed=True).hex()
            # print("Public Key", public_key)
            self._key = key
            self._private_key = key.to_hex()  # == seed
            self._public_key = public_key
        except Exception as e:
            print("Exception {} reading RSA private key".format(e))
        # print("identifier", self.identifier().hex())
        self._address = self.address()

    def identifier(self):
        """Returns double hash of pubkey as per btc standards"""
        return hashlib.new('ripemd160', sha256(bytes.fromhex(self._public_key)).digest()).digest()

    def address(self):
        """Returns properly serialized address from pubkey as per btc standards"""
        vh160 = self.address_version_for_subtype(self._subtype) + self.identifier()  # raw content
        chk = sha256(sha256(vh160).digest()).digest()[:4]
        return base58.b58encode(vh160 + chk).decode('utf-8')

    @classmethod
    def public_key_to_address(cls, public_key: Union[bytes, str],
                              subtype: SignerSubType=SignerSubType.MAINNET_REGULAR) -> str:
        """Reconstruct an address from the public key"""
        if type(public_key) == str:
            identifier = hashlib.new('ripemd160', sha256(bytes.fromhex(public_key)).digest()).digest()
        else:
            identifier = hashlib.new('ripemd160', sha256(public_key).digest()).digest()
        vh160 = cls.address_version_for_subtype(subtype) + identifier  # raw content
        checksum = sha256(sha256(vh160).digest()).digest()[:4]
        return base58.b58encode(vh160 + checksum).decode('utf-8')

    @classmethod
    def verify_signature(cls, signature: Union[bytes, str], public_key: Union[bytes, str], buffer: bytes,
                         address: str='') -> None:
        """Verify signature from raw signature. Address may be used to determine the sig type"""
        raise ValueError("SignerECDSA.verify_signature not impl.")

    @classmethod
    def verify_bis_signature(cls, signature: str, public_key: str, buffer: bytes, address: str = '') -> None:
        """Verify signature from bismuth tx network format (ecdsa sig and pubkey are b64 encoded)
        Returns None, but raises ValueError if needed."""
        public_key = b64decode(public_key)
        valid = verify_signature(b64decode(signature), buffer, public_key)
        if not valid:
            raise ValueError(f"Invalid signature from {address}")
        # Reconstruct address from pubkey to make sure it matches
        if address != cls.public_key_to_address(public_key):
            raise ValueError("Attempt to spend from a wrong address")

    @classmethod
    def verify_bis_signature_raw(cls, signature: bytes, public_key: bytes, buffer: bytes, address: str = '') -> None:
        """Verify signature from bin format
        Returns None, but raises ValueError if needed."""
        # print("ecdsa verify_bis_signature_raw pubkey", public_key, "sig", signature)
        valid = verify_signature(signature, buffer, public_key)
        if not valid:
            raise ValueError(f"Invalid signature from {address}")
        # Reconstruct address from pubkey to make sure it matches
        if address != cls.public_key_to_address(public_key):
            raise ValueError("Attempt to spend from a wrong address")

    def sign_buffer_raw(self, buffer: bytes) -> bytes:
        """Sign a buffer, sends a raw bytes array"""
        # TODO: see "custom_nonce" optional item
        return self._key.sign(buffer)

    def sign_buffer_for_bis(self, buffer: bytes) -> str:
        """Sign a buffer, sends under the format expected by bismuth network format"""
        return b64encode(self.sign_buffer_raw(buffer)).decode('utf-8')




from time import time
from coincurve import verify_signature
import ed25519
from base64 import b64decode
from Cryptodome.Hash import SHA
from Cryptodome.PublicKey import RSA
from Cryptodome.Signature import PKCS1_v1_5
import nacl
import nacl.signing

NB = 1000

ECDSA = {
    "buffer": "('1559472321.00', 'Bis1SAk19HCWpDAThwFiaP9xA6zWjzsga7Hog', 'f6c0363ca1c5aa28cc584252e65a63998493ff0a5ec1bb16beda9bac', '0.63000000', '', 'fake_tx_info')",
    "signature": "MEQCIHu5eNOhmebPTxqHWEMsLJcvFa8i+8hMrsGVXt5xI7rmAiAwFbs3HSGnKiujcbbzRAkVmgTlbG2vU3kOyGe9YWTqfA==",
    "public_key": "A0l0a6ARznKt6nWOCSFZYiuqpQCfrKcq1DFnktgot3lq",
}

ED25519 = {
    "buffer": "('1557384490.32', 'Bis13AbAZwMeY1C5GuFuVuVKLSjr3RdKG63g4CEx6epwSbhpuDU3rj', 'f6c0363ca1c5aa28cc584252e65a63998493ff0a5ec1bb16beda9bac', '0.63443243', '', 'fake_tx_info')",
    "signature": "A9xXtvrwGUSM6k50OVJ9zmx7X+yYA8FGVxyNJ9FPvLYjWKYkJwLGT093XO1X2i8kSh92Q/JVpMQK1AlMaqgRBQ==",
    "public_key": "5YsUhytlWuyWCqIYr1PeXBpTOVWH+WEPnwSL7hfbaAo=",
}

RSA_ = {
    "buffer": "('1557384490.32', 'd35d8ddd4ded7074e5bf092cb716dd4d42d44a2c615115113a6f96dc', 'f6c0363ca1c5aa28cc584252e65a63998493ff0a5ec1bb16beda9bac', '0.63443243', '', 'fake_tx_info')",
    "signature": "iSds5K/vWHSjljBexucIiEWqO48nDK1PN27LvkgptKB4cCQ5EVUOLV8n4JoMjmWIDFs+xrJUwyLerVxcgtP0IHxajHPyfZP3WX9B0YT7bymXIPHWwpTBzDM3gT2GMyMr+GrUWKpwhN0T1nzHxvJ5EoEZQ5FvxD66bDFASZtrB7ABggsSaS4BXB27KQQexJWnQphaMWLdEhlFN1JQW+mlLlMijh+iM9ZFO1i0b2e2otzwAmB/rd+qblDymZsqVcIB72bUvMq/ZGLYGPy/2FbnYycOX1GzqYb0xIjBrKopFu2bBGnL1Fcsw84MHOEiu0XC1xx9wZLGIms902qr+SQky1XBU2DVR2oBPdMZcIRRe2WNUFVB7ybh8M7u8i3YLP4KutKpFgyg7uJbC33BzyzeK1KxIuvwjb9gVSRwjjDFg9GNwvtCHnBpff4ROyL+nPrH8/wRfxapRfCndE6w8WMn4eAofh80VjPnmUiGGZaCrH4lLlBEfF87efn/Cw+gOJCkimVHhNi6J+J2K2ynJbIbn4yUK93yXTRuSYeRj48HdbPZ1Ni3hyyvrPmnnec7AcKHA6TlrJOhUpW8DxAGU1hNazU/WYNt/WM37ofpO9M/mfKtpgyh0Dk0HCnFQfXFCQxQg5fCi8gfUNSJsZEaSR/aJQmgyHVRrWSV2Rw90ZjLDKE=",
    "public_key": "LS0tLS1CRUdJTiBQVUJMSUMgS0VZLS0tLS0KTUlJQ0lqQU5CZ2txaGtpRzl3MEJBUUVGQUFPQ0FnOEFNSUlDQ2dLQ0FnRUFuN2U5dTBIbjZhN3VmdUZDZ2lIcApmem8vWng2amdZb2RLeWZhTm5SS3JhNWsyYm9oVHB2cWRWdVkwRzZGTjNUUVFHbzRXWE5PdFV2c1BRYmQ5REJMCnozd3NyVTY0dWFNVUlNdUZHWnZYYjJuelZKeGlhY3V1S2ZQa01HM1E4RFk4WlhCSWFxNmNibzFBR3RPMy9qcTYKNXNuSm9kR0F5VUFuTGFnK21oSWdTZ3dPOXRqeEs4M3pSQXRTdW1EWnZWTmNzM2liMS9rcm5ydkZjcUFNb0Z6NQpJVlF1Wlh2MUpLYzlvTXRIUCsxak5rWU9rN3BjR0NxVDVXYVJ0LzRWSGI5dmdiQnB3SXVBY09qS0w5b1lVS3dKCkxFZDl3SmFHMWVGSEIrbkxqdmIzbmtVa2phTDdPTGFYMGdBYjNmajFEdUwyeFNSd0hRYVVZOTJlOHlPbXRXN0EKU3FPM2UralJadGtycVVGMkxMWjlucWsvUllTMWV5UWJSWC8xMTJsdnhpK3RVbStjbDNxZER3Y2dzZTA5OVJOZAo0NGlXYkZYaFcwVFRPZ3ptMXhnVmR5SFJib3k4dktoTnhHTFIxUFYwQmMvZm5aZ2RtNSs1YkhkNWZEc2Y4R2kvCmpNYnRreGFwY09MNFB6WW9MZWE4dGcxRmFYZTZSOHhzS1pRakdKNkFZNEg0Y3RzU1M4bkVGQ0pGNjJLWkRyTngKK1hFZ01XS0xvRDJrRVF1aWhqa2t2ZGJDb0t5RE5NbjQ4UExBc2dCQnl6ZHFRVkUxNE5KNklBQWtwdmovSXVwRQpKYnl5cXEzaGZiRUhKWnBXeUtjM2lVa0RibC9tK1BCdEdVYm43UDFEWWxWTGV4VDM4dldiN0tLeFg1TEREVUxkCnZqRnBQb3h4Y1QvcWxFbTQ2cDB4dk04Q0F3RUFBUT09Ci0tLS0tRU5EIFBVQkxJQyBLRVktLS0tLQ==",
}


def verify_ecdsa():
    sig = b64decode(ECDSA["signature"])
    buffer = ECDSA["buffer"].encode()
    public_key = b64decode(ECDSA["public_key"])
    for i in range(NB):
        valid = verify_signature(sig, buffer, public_key)
    return valid


def verify_ed25519():
    sig = b64decode(ED25519["signature"])
    buffer = ED25519["buffer"].encode()
    public_key = b64decode(ED25519["public_key"])
    verifying_key = ed25519.VerifyingKey(public_key)
    for i in range(NB):
        verifying_key.verify(sig, buffer)


def verify_ed25519_nacl():
    sig = b64decode(ED25519["signature"])
    buffer = ED25519["buffer"].encode()
    public_key = b64decode(ED25519["public_key"])

    vk = nacl.signing.VerifyKey(public_key)
    for i in range(NB):
        vk.verify(buffer, sig)


def verify_rsa():
    sig = b64decode(RSA_["signature"])
    buffer = RSA_["buffer"].encode()
    public_key = RSA_["public_key"]
    public_key_pem = b64decode(public_key).decode('utf-8')
    public_key_object = RSA.importKey(public_key_pem)
    verifier = PKCS1_v1_5.new(public_key_object)
    sha_hash = SHA.new(buffer)
    for i in range(NB):
        valid = verifier.verify(sha_hash, sig)
    return valid


def timer(func):
    start = time()
    func()
    delay = time() - start
    print(func, delay, f"{NB/delay:0.0f} tx/s")


if __name__ == "__main__":
    timer(verify_rsa)
    timer(verify_ecdsa)
    timer(verify_ed25519)
    timer(verify_ed25519_nacl)

"""
core i7

<function verify_rsa at 0x7fa9f02330d0> 1.3092164993286133 764 tx/s
<function verify_ecdsa at 0x7fa9f851ce18> 0.07944917678833008 12587 tx/s
<function verify_ed25519 at 0x7fa9f13e3620> 3.6701412200927734 272 tx/s
<function verify_ed25519_nacl at 0x7fa9f13e36a8> 0.09578156471252441 10440 tx/s

"""



"""

"""

import random
from base64 import b64decode, b64encode
from hashlib import sha256
from os import urandom
from typing import Union

import base58
import ed25519
from polysign.signer import Signer, SignerType, SignerSubType


class SignerED25519(Signer):

    __slots__ = ('_key', )

    _address_versions = {SignerSubType.MAINNET_REGULAR: b'\x03\xb8\x6c\xf3',
                         SignerSubType.MAINNET_MULTISIG: b'\x03\xb8\x72\x14',
                         SignerSubType.TESTNET_REGULAR: b'\x11\xc2\xce\x7c',
                         SignerSubType.TESTNET_MULTISIG: b'\x0f\x54\xfd\x2d'}

    def __init__(self, private_key: Union[bytes, str]=b'', public_key: Union[bytes, str]=b'', address: str='',
                 compressed: bool=False, subtype: SignerSubType=SignerSubType.MAINNET_REGULAR):
        super().__init__(private_key, public_key, address, compressed=compressed, subtype=subtype)
        self._key = None
        self._type = SignerType.ED25519

    def from_private_key(self, private_key: Union[bytes, str], subtype: SignerSubType=SignerSubType.MAINNET_REGULAR):
        """Accepts both bytes[32] or str (hex format)"""
        if subtype != SignerSubType.MAINNET_REGULAR:
            self._subtype = subtype
        if type(private_key) == str:
            return self.from_seed(private_key, subtype=self._subtype)
        return self.from_seed(private_key.hex(), subtype=self._subtype)

    def from_full_info(self, private_key: Union[bytes, str], public_key: Union[bytes, str]=b'', address: str='',
                       subtype: SignerSubType = SignerSubType.MAINNET_REGULAR, verify: bool=True):
        raise ValueError("SignerED25519.from_full_info not impl.")

    def from_seed(self, seed: str='', subtype: SignerSubType=SignerSubType.MAINNET_REGULAR):
        """Creates key from seed - for ED25519, seed = pk - 32 bytes random buffer"""
        if subtype != SignerSubType.MAINNET_REGULAR:
            self._subtype = subtype
        if len(seed) > 64:
            # Too long seed, trim (could use better scheme for more entropy)
            seed = seed[:64]
        elif seed == '':
            # No seed, use urandom
            seed = urandom(32)
        elif len(seed) < 64:
            # Too short seed, use as PRNG seed
            random.seed(seed)
            seed = hex(random.getrandbits(32*8))[2:]
            while len(seed) < 64:
                seed = '0' + seed
            assert len(seed) == 64
        try:
            # print("SEED", seed)
            # TODO: check flow, there may be many unnecessary hex-byte-hex-bytes conversions from top to bottom
            key = ed25519.SigningKey(bytes.fromhex(seed))
            hexa = key.to_ascii(encoding="hex").decode('utf-8')
            # print("ED25519 Privk Key", hexa)  # e5b42f3c-3fe02e16-1d42ff47-07a174a5 715b2badc7d4d3aebbea9081bd9123d5
            verifying_key = key.get_verifying_key()
            public_key = verifying_key.to_ascii(encoding="hex").decode('utf-8')
            # public_key = hexa[32:]
            # print("ED25519 Public Key", public_key)
            self._key = key
            self._private_key = hexa
            self._public_key = public_key
        except Exception as e:
            print("Exception {} reading ED25519 private key".format(e))
        # print("identifier", self.identifier().hex())
        self._address = self.address()

    """
    def identifier(self):
        #Returns double hash of pubkey as per btc standards
        return hashlib.new('ripemd160', sha256(bytes.fromhex(self._public_key)).digest()).digest()
    """

    def address(self) -> str:
        """Returns properly serialized address from pubkey"""
        # No double hash for pubkey, nor for checksum
        base = self.address_version_for_subtype(self._subtype) + bytes.fromhex(self._public_key)  # raw content
        chk = sha256(base).digest()[:4]
        return base58.b58encode(base + chk).decode('utf-8')

    @classmethod
    def public_key_to_address(cls, public_key: Union[bytes, str],
                              subtype: SignerSubType=SignerSubType.MAINNET_REGULAR) -> str:
        """Reconstruct an address from the public key"""
        # TODO: same for this family, could factorize in an ancestor with other methods
        if type(public_key) == str:
            public_key = bytes.fromhex(public_key)
        base = cls.address_version_for_subtype(subtype) + public_key  # raw content
        checksum = sha256(base).digest()[:4]
        return base58.b58encode(base + checksum).decode('utf-8')

    @classmethod
    def verify_signature(cls, signature: Union[bytes, str], public_key: Union[bytes, str], buffer: bytes,
                         address: str='') -> None:
        """Verify signature from raw signature. Address may be used to determine the sig subtype"""
        try:
            # print("verif", signature, public_key, len(public_key))
            verifying_key = ed25519.VerifyingKey(public_key)
            verifying_key.verify(signature, buffer)
        except Exception as e:
            print(e)
            raise ValueError(f"Invalid ED25519 signature from {address}")
        # Reconstruct address from pubkey to make sure it matches
        address_rebuild = cls.public_key_to_address(public_key)
        if address != address_rebuild:
            raise ValueError(f"Attempt to spend from a wrong address {address} instead of {address_rebuild}")

    @classmethod
    def verify_bis_signature(cls, signature: str, public_key: str, buffer: bytes, address: str = '') -> None:
        """Verify signature from bismuth tx network format (ecdsa sig and pubkey are b64 encoded)
        Returns None, but raises ValueError if needed."""
        cls.verify_signature(b64decode(signature), b64decode(public_key), buffer, address)

    @classmethod
    def verify_bis_signature_raw(cls, signature: bytes, public_key: bytes, buffer: bytes, address: str = '') -> None:
        """Verify signature from bin format
        Returns None, but raises ValueError if needed."""
        cls.verify_signature(signature, public_key, buffer, address)

    def sign_buffer_raw(self, buffer: bytes) -> bytes:
        """Sign a buffer, sends a raw bytes array"""
        return self._key.sign(buffer)

    def sign_buffer_for_bis(self, buffer: bytes) -> str:
        """Sign a buffer, sends under the format expected by bismuth network format"""
        return b64encode(self.sign_buffer_raw(buffer)).decode('utf-8')



import base58
from hashlib import sha256

ADDRESS_VERSION = b"\x01\x75\x07"

old_addr = "15bj2HB2UbmjEZgXyEW4M8MhUL5TXGCN8L"
decoded = base58.b58decode(old_addr).hex()
print("decoded", decoded)
# 00 3271b95d0a3af5b1095502895a422a907b3084129e808f85
# 80 3271b95d0a3af5b1095502895a422a907b308412 9e808f85
# 1 byte network, 20 bytes PubK, 4 bytes checksum
pubk = decoded[2:42]
print("pubk", pubk)

vh160 = ADDRESS_VERSION + bytes.fromhex(pubk)
chk = sha256(sha256(vh160).digest()).digest()[:4]
print("chk", chk)
new = base58.b58encode(vh160 + chk).decode("utf-8")
print("new", new)
assert new == "CRWGZiLqSUbCTWkm3ABp5qpXdun2h8DJCKYF"



"""
Check test vectors with all bis addresses schemes
"""

import json
import pytest
import sys

sys.path.append("../")
from polysign.signerfactory import SignerFactory
from polysign.signer import SignerType, SignerSubType


def test_bis_vectors(verbose: bool=False):
    with open("./vectors/bis_vectors.json") as f:
        wallets = json.load(f)
    for wallet in wallets:
        private_key = wallet["private_key"]
        signer_type = SignerType.RSA
        if wallet["type"] == "ECDSA":
            signer_type = SignerType.ECDSA
        if wallet["type"] == "ED25519":
            signer_type = SignerType.ED25519
        signer_subtype = SignerSubType.MAINNET_REGULAR
        if wallet["sub_type"] == "TESTNET_REGULAR":
            signer_subtype = SignerSubType.TESTNET_REGULAR
        signer = SignerFactory.from_private_key(private_key, signer_type=signer_type, subtype=signer_subtype)
        signer_dict = signer.to_dict()
        if verbose:
            print(signer_dict)
        assert signer_dict["address"] == wallet["address"]
        assert signer_dict["public_key"] == wallet["public_key"]
        assert signer_dict["type"] == wallet["type"]
        assert signer_dict["sub_type"] == wallet["sub_type"]
        # In case privkey was submitted as uppercase
        assert signer_dict["private_key"].lower() == wallet["private_key"].lower()


if __name__ == "__main__":
    test_bis_vectors(True)



"""
Generates test vectors for ECDSA signatures, from given timestamp
"""

import json
import base64
import random
import sys
from time import time
from decimal import *

sys.path.append("../")
from polysign.signer import SignerType
from polysign.signerfactory import SignerFactory

TIMESTAMP = 1559472321.17


def quantize_two(value):
    value = Decimal(value)
    value = value.quantize(Decimal("0.00"))
    return value


def quantize_eight(value):
    value = Decimal(value)
    value = value.quantize(Decimal("0.00000000"))
    return value


def get_tx_vector(i: int, a_tx: dict) -> dict:
    tx = dict(a_tx)  # deep copy
    tx["timestamp"] = "%.2f" % quantize_two(tx["timestamp"])
    tx["amount"] = "%.8f" % quantize_eight(tx["amount"])
    vector = {"id": i, "tx": tx}
    # adjust and recreate the buffer
    buffer = str(
        (
            tx["timestamp"],
            tx["address"],
            tx["recipient"],
            tx["amount"],
            tx["operation"],
            tx["openfield"],
        )
    ).encode("utf-8")
    vector["buffer"] = buffer.decode("utf-8")
    print("\nBuffer {}: {}".format(i, vector["buffer"]))
    # Sign the tx
    signature = signer.sign_buffer_for_bis(buffer)
    print("ECDSA Signature", signature)
    vector["signature"] = signature

    public_key_hex = signer.to_dict()["public_key"]
    public_key_net = base64.b64encode(bytes.fromhex(public_key_hex)).decode("utf-8")
    vector["public_key"] = public_key_net

    full_tx = [
        tx["timestamp"],
        tx["address"],
        tx["recipient"],
        tx["amount"],
        signature,
        public_key_net,
        tx["operation"],
        tx["openfield"],
    ]
    vector["websocket_command"] = '["mpinsert", [{}]]'.format(json.dumps(full_tx))
    print("websocket : {}".format(vector["websocket_command"]))
    SignerFactory.verify_bis_signature(
        signature, public_key_net, buffer, "Bis1SAk19HCWpDAThwFiaP9xA6zWjzsga7Hog"
    )
    return vector


if __name__ == "__main__":

    # sign the TX data with our ecdsa test address
    signer = SignerFactory.from_private_key(
        "e5b42f3c3fe02e161d42ff4707a174a5715b2badc7d4d3aebbea9081bd9123d5",
        SignerType.ECDSA,
    )
    print(signer.to_dict())
    assert signer.to_dict()["address"] == "Bis1SAk19HCWpDAThwFiaP9xA6zWjzsga7Hog"

    current_time = TIMESTAMP
    vectors = []
    tx = {
        "timestamp": int(current_time),  # 2 decimals will be added to conform.
        "address": "Bis1SAk19HCWpDAThwFiaP9xA6zWjzsga7Hog",
        "recipient": "f6c0363ca1c5aa28cc584252e65a63998493ff0a5ec1bb16beda9bac",
        "amount": "0.63",  # Decimal will be added
        "operation": "",
        "openfield": "fake_tx_info",
    }
    vectors.append(get_tx_vector(0, tx))

    random.seed(555)  # reproducible results
    tx["openfield"] = ""
    for i in range(10):
        tx["timestamp"] = current_time + i / 100
        tx["amount"] = random.randint(0, 1000) / 1000
        vectors.append(get_tx_vector(i + 1, tx))
    # print(vectors)
    with open("ecdsa_txs.json", "w") as f:
        json.dump(vectors, f, indent=2)



"""
Test precalc vector of ecdsa sigs
"""

import json
import base64
import pytest
import sys

sys.path.append("../")
from polysign.signerfactory import SignerFactory
from polysign.signer import SignerType


def test_ecdsa_txs(verbose: bool=False):
    signer = SignerFactory.from_private_key(
        "e5b42f3c3fe02e161d42ff4707a174a5715b2badc7d4d3aebbea9081bd9123d5",
        SignerType.ECDSA,
    )
    if verbose:
        print(signer.to_dict())
    assert signer.to_dict()["address"] == "Bis1SAk19HCWpDAThwFiaP9xA6zWjzsga7Hog"

    public_key_hex = signer.to_dict()["public_key"]
    public_key_net = base64.b64encode(bytes.fromhex(public_key_hex)).decode("utf-8")

    with open("./vectors/ecdsa_txs.json", "r") as f:
        txs = json.load(f)
        assert txs[0]["public_key"] == public_key_net
        for tx in txs:
            if verbose:
                print(f"Testing tx #{tx['id']}")
            buffer = tx['buffer'].encode()
            # Recalc sig from buffer
            signature = signer.sign_buffer_for_bis(buffer)
            # Make sure sig is same as before
            assert signature == tx["signature"]
            # Make sure it still verifies as well
            SignerFactory.verify_bis_signature(
                signature, public_key_net, buffer, "Bis1SAk19HCWpDAThwFiaP9xA6zWjzsga7Hog"
            )


if __name__ == "__main__":
    test_ecdsa_txs(True)



"""
Test - Generate vanity ecdsa address
"""

import sys
from os import urandom
from time import time
from secrets import token_hex
from multiprocessing import Process

sys.path.append("../")
from polysign.signer import SignerType, SignerSubType
from polysign.signerfactory import SignerFactory

# TODO: click would be a better choice than tornado. I'm lazy.
from tornado.options import define, options


define("processes", default=4, help="Process count to start (default 4)", type=int)
define("string", help="String to find in the address", type=str)
define("case", default=False, help="be case sensitive (default false)", type=bool)
define("max", default=100, help="max hit per process (default 100)", type=int)

alphabet = '123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'


def find_it(string: str):
    found = 0
    while True:
        # pk = urandom(32).hex()
        pk = token_hex(32)  # +50% time, but supposed to be cryptographically secure
        signer = SignerFactory.from_seed(pk, signer_type, subtype=subtype)
        address = signer._address
        if not options.case:
            address = address.lower()
        if string in address:
            print(signer._address, pk)
            found += 1
            if found > options.max:
                return


if __name__ == "__main__":
    options.parse_command_line()
    # TODO: could be given on command line
    signer_type = SignerType.ECDSA
    subtype = SignerSubType.MAINNET_REGULAR

    print("Looking for '{}'".format(options.string))
    if not options.case:
        options.string = options.string.lower()
        print("Case InsEnsITivE")
    else:
        print("Case sensitive")
    # Check charset is ok ? not easy with case insensitive...

    processes = []
    for i in range(options.processes):
        p = Process(target=find_it, args=(options.string,))
        p.start()
        processes.append(p)

    for p in processes:
        p.join()
    # find_it(options.string)
    """
    start = time()
    for i in range(100):
        # pk = urandom(32).hex()
        pk = token_hex(32)  # +50% time, but supposed to be cryptographically secure
        signer = SignerFactory.from_seed(pk, signer_type, subtype=subtype)
        # print("  ", signer.to_dict())
        print(signer.to_dict()["address"], pk)
    print()
    print(time() - start)
    """



"""

"""

import base58
import hashlib
import random
from os import urandom
from polysign.signer import Signer, SignerType, SignerSubType
from typing import Union
from hashlib import sha256
from base64 import b64decode, b64encode
from coincurve import PrivateKey


class SignerBTC(Signer):

    __slots__ = ('_key', )

    _address_versions = {SignerSubType.MAINNET_REGULAR: b'\x00', SignerSubType.TESTNET_REGULAR: b'\x6f'}

    def __init__(self, private_key: Union[bytes, str]=b'', public_key: Union[bytes, str]=b'', address: str='',
                 compressed: bool=True, subtype: SignerSubType=SignerSubType.MAINNET_REGULAR):
        super().__init__(private_key, public_key, address, compressed, subtype=subtype)
        self._key = None
        self._type = SignerType.ECDSA

    def from_private_key(self, private_key: Union[bytes, str], subtype: SignerSubType=SignerSubType.MAINNET_REGULAR):
        """Accepts both bytes[32] or str (hex format)"""
        if type(private_key) == str:
            return self.from_seed(private_key)
        return self.from_seed(private_key.hex())

    def from_full_info(self, private_key: Union[bytes, str], public_key: Union[bytes, str]=b'', address: str='',
                       subtype: SignerSubType = SignerSubType.MAINNET_REGULAR, verify: bool=True):
        print('TODO - SignerBTC.from_full_info')

    def from_seed(self, seed: str='', subtype: SignerSubType=SignerSubType.MAINNET_REGULAR):
        """Creates key from seed - for ecdsa, seed = pk - 32 bytes random buffer"""
        if subtype != SignerSubType.MAINNET_REGULAR:
            self._subtype = subtype
        if len(seed) > 64:
            # Too long seed, trim (could use better scheme for more entropy)
            seed = seed[:64]
        elif seed == '':
            # No seed, use urandom
            seed = urandom(32)
        elif len(seed) < 64:
            # Too short seed, use as PRNG seed
            random.seed(seed)
            seed = random.getrandbits(32*8).hex()
        try:
            key = PrivateKey.from_hex(seed)
            public_key = key.public_key.format(compressed=True).hex()
            # print("Public Key", public_key)
            self._key = key
            self._private_key = key.to_hex()  # == seed
            self._public_key = public_key
        except Exception as e:
            print("Exception {} reading RSA private key".format(e))
        # print("identifier", self.identifier().hex())
        self._address = self.address()

    def identifier(self):
        """Returns double hash of pubkey as per btc standards"""
        return hashlib.new('ripemd160', sha256(bytes.fromhex(self._public_key)).digest()).digest()

    def address(self):
        """Returns properly serialized address from pubkey as per btc standards"""
        vh160 = self.address_version_for_subtype(self._subtype) + self.identifier()  # raw content
        chk = sha256(sha256(vh160).digest()).digest()[:4]
        return base58.b58encode(vh160 + chk).decode('utf-8')

    @classmethod
    def public_key_to_address(cls, public_key: Union[bytes, str],
                              subtype: SignerSubType=SignerSubType.MAINNET_REGULAR) -> str:
        """Reconstruct an address from the public key"""
        if type(public_key) == str:
            identifier = hashlib.new('ripemd160', sha256(bytes.fromhex(public_key)).digest()).digest()
        else:
            identifier = hashlib.new('ripemd160', sha256(public_key).digest()).digest()
        vh160 = cls.address_version_for_subtype(subtype) + identifier  # raw content
        checksum = sha256(sha256(vh160).digest()).digest()[:4]
        return base58.b58encode(vh160 + checksum).decode('utf-8')

    @classmethod
    def verify_signature(cls, signature: Union[bytes, str], public_key: Union[bytes, str], buffer: bytes,
                         address: str='') -> None:
        """Verify signature from raw signature. Address may be used to determine the sig type"""
        raise ValueError("SignerBTC.verify_signature not impl.")

    @classmethod
    def verify_bis_signature(cls, signature: str, public_key: str, buffer: bytes, address: str = '') -> None:
        """Verify signature from bismuth tx network format (ecdsa sig and pubkey are b64 encoded)
        Returns None, but raises ValueError if needed."""
        public_key = b64decode(public_key).decode('utf-8')
        # print(public_key)

        """ TODO
        public_key_object = RSA.importKey(public_key_pem)
        signature_decoded = b64decode(signature)
        verifier = PKCS1_v1_5.new(public_key_object)
        sha_hash = SHA.new(buffer)
        if not verifier.verify(sha_hash, signature_decoded):
            raise ValueError(f"Invalid signature from {address}")
        """
        # Reconstruct address from pubkey to make sure it matches
        if address != cls.public_key_to_address(public_key):
            raise ValueError("Attempt to spend from a wrong address")

    def sign_buffer_raw(self, buffer: bytes) -> bytes:
        """Sign a buffer, sends a raw bytes array"""
        pass

    def sign_buffer_for_bis(self, buffer: bytes) -> str:
        """Sign a buffer, sends under the format expected by bismuth network format"""
        # sig is b64 encoded
        return b64encode(self.sign_buffer_raw(buffer))



"""
Tests address generation from private keys, with very various schemes including other existing cryptos test vectors
"""

import json
import base58
import pytest
import sys

sys.path.append("../")
from polysign.signer import SignerType, SignerSubType
from polysign.signerfactory import SignerFactory


# Never to be used for real addresses - from https://en.bitcoin.it/wiki/BIP_0032_TestVectors
TEST_SEED = "e8f32e723decf4051aefac8e2c93c9c5b214313817cdb01a1494b917c8436b35"

TEST_PK = "e5b42f3c3fe02e161d42ff4707a174a5715b2badc7d4d3aebbea9081bd9123d5"


def test_rsa_address1(verbose: bool=False):
    # RSA Test
    with open("./vectors/rsa_wallet1.json") as f:
        wallet = json.load(f)
    rsa_private_key = wallet["Private Key"]
    signer = SignerFactory.from_private_key(rsa_private_key, SignerType.RSA)
    if verbose:
        print(signer.to_json())
    signer_dict = signer.to_dict()
    assert signer_dict["address"] == wallet["Address"]
    assert signer_dict["public_key"] == "-----BEGIN PUBLIC KEY-----\n" \
                                        "MIICIjANBgkqhkiG9w0BAQEFAAOCAg8AMIICCgKCAgEAn7e9u0Hn6a7ufuFCgiHp\n" \
                                        "fzo/Zx6jgYodKyfaNnRKra5k2bohTpvqdVuY0G6FN3TQQGo4WXNOtUvsPQbd9DBL\n" \
                                        "z3wsrU64uaMUIMuFGZvXb2nzVJxiacuuKfPkMG3Q8DY8ZXBIaq6cbo1AGtO3/jq6\n" \
                                        "5snJodGAyUAnLag+mhIgSgwO9tjxK83zRAtSumDZvVNcs3ib1/krnrvFcqAMoFz5\n" \
                                        "IVQuZXv1JKc9oMtHP+1jNkYOk7pcGCqT5WaRt/4VHb9vgbBpwIuAcOjKL9oYUKwJ\n" \
                                        "LEd9wJaG1eFHB+nLjvb3nkUkjaL7OLaX0gAb3fj1DuL2xSRwHQaUY92e8yOmtW7A\n" \
                                        "SqO3e+jRZtkrqUF2LLZ9nqk/RYS1eyQbRX/112lvxi+tUm+cl3qdDwcgse099RNd\n" \
                                        "44iWbFXhW0TTOgzm1xgVdyHRboy8vKhNxGLR1PV0Bc/fnZgdm5+5bHd5fDsf8Gi/\n" \
                                        "jMbtkxapcOL4PzYoLea8tg1FaXe6R8xsKZQjGJ6AY4H4ctsSS8nEFCJF62KZDrNx\n" \
                                        "+XEgMWKLoD2kEQuihjkkvdbCoKyDNMn48PLAsgBByzdqQVE14NJ6IAAkpvj/IupE\n" \
                                        "Jbyyqq3hfbEHJZpWyKc3iUkDbl/m+PBtGUbn7P1DYlVLexT38vWb7KKxX5LDDULd\n" \
                                        "vjFpPoxxcT/qlEm46p0xvM8CAwEAAQ==\n" \
                                        "-----END PUBLIC KEY-----"
    assert signer_dict["type"] == "RSA"
    assert signer_dict["sub_type"] == "MAINNET_REGULAR"
    assert signer_dict["private_key"] == rsa_private_key


def test_btc_address(verbose: bool = False):
    # BTC ECDSA Test - seed is in fact a 32 byte privkey (random, no constraint)
    signer = SignerFactory.from_seed(TEST_SEED, SignerType.BTC)
    signer_dict = signer.to_dict()
    if verbose:
        print(signer_dict)
    assert signer.to_dict()["address"] == "15mKKb2eos1hWa6tisdPwwDC1a5J1y9nma"
    assert signer_dict["public_key"] == "0339a36013301597daef41fbe593a02cc513d0b55527ec2df1050e2e8ff49c85c2"
    assert signer_dict["private_key"] == "e8f32e723decf4051aefac8e2c93c9c5b214313817cdb01a1494b917c8436b35"
    assert signer_dict["sub_type"] == "MAINNET_REGULAR"
    assert signer_dict["type"] == "ECDSA"


def test_crw_address(verbose: bool = False):
    # From https://gitlab.crown.tech/crown/crown-core/issues/210
    wif = "5KZT4KD9mFd45h44LkR2ibUdt5EMF15YBtBxuRtomWAFgsAa2wa"
    wifdecode = base58.b58decode(wif).hex()
    if verbose:
        print("wifdecode", wifdecode)
    # 80e5b42f3c3fe02e161d42ff4707a174a5715b2badc7d4d3aebbea9081bd9123d566129939
    # 80 e5b42f3c3fe02e161d42ff4707a174a5715b2badc7d4d3aebbea9081bd9123d5 66129939
    # 1 byte network, 32 bytes PK, 4 bytes checksum
    # CRW seems to use the uncompressed pubaddress format, btc uses compressed one.
    # fits with https://crown.tech/paper_wallet/index.html
    pk = wifdecode[2:66]
    if verbose:
        print("pk", pk)
    signer = SignerFactory.from_seed(pk, SignerType.CRW)
    signer_dict = signer.to_dict()
    if verbose:
        print(signer_dict)
    assert signer.to_dict()["address"] == "CRWGg6VvaNe6zhQ46wuEBci8VerP4qVTw8qq"
    assert signer_dict["public_key"] == "0449746ba011ce72adea758e092159622baaa5009faca72ad4316792d828b7796a738dbf2ce8386d9b5aa1275c70d160e774dc175f41e005311616bf13942f57ad"
    assert signer_dict["private_key"] == "e5b42f3c3fe02e161d42ff4707a174a5715b2badc7d4d3aebbea9081bd9123d5"
    assert signer_dict["sub_type"] == "MAINNET_REGULAR"
    assert signer_dict["type"] == "ECDSA"
    assert signer_dict["compressed"] == False


def test_bis_mainnet(verbose: bool = False):
    signer = SignerFactory.from_seed(TEST_PK, SignerType.ECDSA)
    signer_dict = signer.to_dict()
    if verbose:
        print(signer_dict)
    assert signer.to_dict()["address"] == "Bis1SAk19HCWpDAThwFiaP9xA6zWjzsga7Hog"
    assert signer_dict["public_key"] == "0349746ba011ce72adea758e092159622baaa5009faca72ad4316792d828b7796a"
    assert signer_dict["private_key"] == "e5b42f3c3fe02e161d42ff4707a174a5715b2badc7d4d3aebbea9081bd9123d5"
    assert signer_dict["sub_type"] == "MAINNET_REGULAR"
    assert signer_dict["type"] == "ECDSA"
    assert signer_dict["compressed"] == True


def test_bis_testnet(verbose: bool = False):
    signer = SignerFactory.from_seed(TEST_PK, SignerType.ECDSA, subtype=SignerSubType.TESTNET_REGULAR)
    signer_dict = signer.to_dict()
    if verbose:
        print(signer_dict)
    assert signer.to_dict()["address"] == "tBisH4YbEridY7Xw2Dv5sP727VA7tV2j5fTk7"
    assert signer_dict["public_key"] == "0349746ba011ce72adea758e092159622baaa5009faca72ad4316792d828b7796a"
    assert signer_dict["private_key"] == "e5b42f3c3fe02e161d42ff4707a174a5715b2badc7d4d3aebbea9081bd9123d5"
    assert signer_dict["sub_type"] == "TESTNET_REGULAR"
    assert signer_dict["type"] == "ECDSA"
    assert signer_dict["compressed"] == True


if __name__ == "__main__":
    test_rsa_address1(True)
    test_btc_address(True)
    test_crw_address(True)
    test_bis_mainnet(True)
    test_bis_testnet(True)



"""
Test - Generate several random ECDSA Bis addresses of different types
"""

import sys
from os import urandom

sys.path.append("../")
from polysign.signer import SignerType, SignerSubType
from polysign.signerfactory import SignerFactory


if __name__ == "__main__":
    for signer_type in [SignerType.ECDSA, SignerType.ED25519]:
        print(signer_type.name)
        for subtype in SignerSubType:
            print(" -", subtype.name)
            for i in range(10):
                pk = urandom(32).hex()
                signer = SignerFactory.from_seed(pk, signer_type, subtype=subtype)
                print("  ", signer.to_dict())
                print(len(signer.to_dict()["address"]))




import json
import re
from base64 import b64encode, b64decode
from hashlib import sha224
from typing import Union

from Cryptodome.Hash import SHA
from Cryptodome.PublicKey import RSA
from Cryptodome.Signature import PKCS1_v1_5

from polysign.signer import Signer, SignerType, SignerSubType

# Compile once and for all
PEM_BEGIN = re.compile(r"\s*-----BEGIN (.*)-----\s+")
PEM_END = re.compile(r"-----END (.*)-----\s*$")


class SignerRSA(Signer):

    __slots__ = ('_key', )

    def __init__(self, private_key: Union[bytes, str]=b'', public_key: Union[bytes, str]=b'', address: str='',
                 compressed: bool = True, subtype: SignerSubType=SignerSubType.MAINNET_REGULAR):
        super().__init__(private_key, public_key, address, compressed=compressed, subtype=subtype)
        # RSA does not have compressed format
        self._type = SignerType.RSA
        # For the Key object
        self._key = None

    @classmethod
    def validate_pem(cls, pem_data: str) -> None:
        """ Validate PEM data
        returns None, raise on error.
        """
        # verify pem as cryptodome does
        match = PEM_BEGIN.match(pem_data)
        if not match:
            raise ValueError("Not a valid PEM pre boundary")
        marker = match.group(1)
        match = PEM_END.search(pem_data)
        if not match or match.group(1) != marker:
            raise ValueError("Not a valid PEM post boundary")
            # verify pem as cryptodome does

    @classmethod
    def public_key_to_address(cls, public_key: Union[bytes, str],
                              subtype: SignerSubType=SignerSubType.MAINNET_REGULAR) -> str:
        """Reconstruct an address from the public key"""
        if type(public_key) != str:
            # But union annotation kept for common interface sake.
            raise ValueError("RSA pubkey are str, pem format")
        return sha224(public_key.encode('utf-8')).hexdigest()

    @classmethod
    def public_key_bin_to_address(cls, public_key: bytes, subtype: SignerSubType=SignerSubType.MAINNET_REGULAR) -> str:
        """Reconstruct an address from the public key"""
        if type(public_key) != bytes:
            # But union annotation kept for common interface sake.
            raise ValueError("public_key_bin_to_address expect bytes, pem format")
        return sha224(public_key).hexdigest()

    def to_json(self) -> str:
        """for RSA, keys are stored as PEM format, not binary"""
        info = self.to_dict()
        return json.dumps(info)

    def from_private_key(self, private_key: Union[bytes, str],
                         subtype: SignerSubType=SignerSubType.MAINNET_REGULAR) -> None:
        if type(private_key) is not str:
            raise RuntimeError('RSA private key have to be strings')
        if subtype != SignerSubType.MAINNET_REGULAR:
            self._subtype = subtype
        try:
            key = RSA.importKey(private_key)
            public_key_readable = key.publickey().exportKey().decode("utf-8")
            if len(public_key_readable) not in (271, 799):
                raise ValueError("Invalid public key length: {}".format(len(public_key_readable)))
            address = sha224(public_key_readable.encode('utf-8')).hexdigest()
            # If we had no error, we can store
            self._key = key
            self._private_key = private_key
            self._public_key = public_key_readable
            self._address = address
        except Exception as e:
            print("Exception {} reading RSA private key".format(e))

    def from_seed(self, seed: str='', subtype: SignerSubType=SignerSubType.MAINNET_REGULAR) -> None:
        """
        if subtype != SignerSubType.MAINNET_REGULAR:
            self._subtype = subtype
        """
        raise ValueError("SignerRsa.from_seed not impl. - seed {}".format(seed))

    def from_full_info(self, private_key: Union[bytes, str], public_key: Union[bytes, str]=b'', address: str='',
                       subtype: SignerSubType = SignerSubType.MAINNET_REGULAR, verify: bool=True):
        raise ValueError("SignerRsa.from_full_info not impl.")

    @classmethod
    def verify_signature(cls, signature: Union[bytes, str], public_key: Union[bytes, str], buffer: bytes,
                         address: str='') -> None:
        """Verify signature from raw signature. Address may be used to determine the sig type"""
        raise ValueError("SignerRsa.verify_signature not impl.")

    @classmethod
    def verify_bis_signature(cls, signature: str, public_key: str, buffer: bytes, address: str = '') -> None:
        """Verify signature from bismuth tx network format (rsa sig is b64 encoded twice)
        Returns None, but raises ValueError if needed."""
        public_key_pem = b64decode(public_key).decode('utf-8')
        # Will raise if does not match
        cls.validate_pem(public_key_pem)
        public_key_object = RSA.importKey(public_key_pem)
        signature_decoded = b64decode(signature)
        verifier = PKCS1_v1_5.new(public_key_object)
        sha_hash = SHA.new(buffer)
        if not verifier.verify(sha_hash, signature_decoded):
            raise ValueError(f"Invalid signature from {address}")
        # Reconstruct address from pubkey to make sure it matches
        if address != cls.public_key_to_address(public_key_pem):
            raise ValueError("Attempt to spend from a wrong address")

    @classmethod
    def normalize_key(cls, s: str) -> str:
        """Re add boundaries and segment by 64 chars wide."""
        chunks = [s[i:i + 64] for i in range(0, len(s), 64)]
        chunks.insert(0, "-----BEGIN PUBLIC KEY-----")
        chunks.append("-----END PUBLIC KEY-----")
        return "\n".join(chunks)

    @classmethod
    def verify_bis_signature_raw(cls, signature: bytes, public_key: bytes, buffer: bytes, address: str = '') -> None:
        """Verify signature from bismuth tx bin stored format (b64 decoded)
        Returns None, but raises ValueError if needed."""
        # Will raise if does not match
        # TODO: Probably be better to use DER (bin) format there for rsa, instead of re-encode to PEM
        # However, if we want to check address... we need the PEM, see below.
        # move to ed25519 would free significant ressources.
        # pem = "-----BEGIN PUBLIC KEY-----\n" + b64encode(public_key).decode() + "\n-----END PUBLIC KEY-----"
        pem = cls.normalize_key(b64encode(public_key).decode())
        cls.validate_pem(pem)
        public_key_object = RSA.importKey(pem)
        verifier = PKCS1_v1_5.new(public_key_object)
        sha_hash = SHA.new(buffer)
        if not verifier.verify(sha_hash, signature):
            raise ValueError(f"Invalid signature from {address}")
        # Reconstruct address from pubkey to make sure it matches
        # print(address, pem, cls.public_key_to_address(pem))
        if address != cls.public_key_to_address(pem):
            raise ValueError("Attempt to spend from a wrong address")

    def sign_buffer_raw(self, buffer: bytes) -> bytes:
        """Sign a buffer, returns a raw bytes array"""
        h = SHA.new(buffer)
        signer = PKCS1_v1_5.new(self._key)
        return signer.sign(h)

    def sign_buffer_for_bis(self, buffer: bytes) -> str:
        """Sign a buffer, sends under the format expected by bismuth network format"""
        # For RSA, sig is b64 encoded
        return b64encode(self.sign_buffer_raw(buffer))



"""
Demoes how to sign a generic tx
"""

import json
import base64
import sys

sys.path.append("../")
from polysign.signer import SignerType
from polysign.signerfactory import SignerFactory


if __name__ == "__main__":
    # Load privkey from our test rsa1.json (do not use IRL!)
    with open("rsa1.json") as f:
        wallet = json.load(f)
    rsa_private_key = wallet["Private Key"]

    # Create a signer. Safer if we can force the Type, but should be ok in the end with just key
    signer = SignerFactory.from_private_key(rsa_private_key, SignerType.RSA)

    # Load our fake tx params
    with open("rsa_tx2.json") as f:
        rsa_tx = json.load(f)
    # Caller is responsible for creating the bin buffer to check the sig against
    buffer = str(
        (
            rsa_tx["timestamp"],
            rsa_tx["address"],
            rsa_tx["recipient"],
            rsa_tx["amount"],
            rsa_tx["operation"],
            rsa_tx["openfield"],
        )
    ).encode("utf-8")
    print("buffer", buffer)

    # Sign the tx
    signature = signer.sign_buffer_for_bis(buffer)
    print("RSA Signature", signature)

    # Extra: recheck (useless, for test only) - Raises if Error
    SignerFactory.verify_bis_signature(
        rsa_tx["signature"], rsa_tx["public_key"], buffer, rsa_tx["address"]
    )
    print("No Error")

    # =================================================================================

    # Now sign the same TX data with our ecdsa test address
    # Could as well be loaded from a wallet, no matter.
    signer = SignerFactory.from_private_key(
        "e5b42f3c3fe02e161d42ff4707a174a5715b2badc7d4d3aebbea9081bd9123d5",
        SignerType.ECDSA,
    )
    print(signer.to_dict())
    assert signer.to_dict()["address"] == "Bis1SAk19HCWpDAThwFiaP9xA6zWjzsga7Hog"

    # adjust and recreate the buffer
    rsa_tx["address"] = "Bis1SAk19HCWpDAThwFiaP9xA6zWjzsga7Hog"
    buffer = str(
        (
            rsa_tx["timestamp"],
            rsa_tx["address"],
            rsa_tx["recipient"],
            rsa_tx["amount"],
            rsa_tx["operation"],
            rsa_tx["openfield"],
        )
    ).encode("utf-8")
    # Sign the tx
    signature = signer.sign_buffer_for_bis(buffer)
    print("ECDSA Signature", signature)

    public_key_hex = signer.to_dict()["public_key"]
    public_key_net = base64.b64encode(bytes.fromhex(public_key_hex)).decode("utf-8")
    print(public_key_hex, public_key_net)

    SignerFactory.verify_bis_signature(
        signature, public_key_net, buffer, "Bis1SAk19HCWpDAThwFiaP9xA6zWjzsga7Hog"
    )
    print("No Error")

    # =================================================================================

    # Now sign the same TX data with our ed25519 test address
    # Could as well be loaded from a wallet, no matter.
    signer = SignerFactory.from_private_key(
        "e5b42f3c3fe02e161d42ff4707a174a5715b2badc7d4d3aebbea9081bd9123d5",
        SignerType.ED25519,
    )
    print(signer.to_dict())
    assert (
        signer.to_dict()["address"]
        == "Bis13AbAZwMeY1C5GuFuVuVKLSjr3RdKG63g4CEx6epwSbhpuDU3rj"
    )

    # adjust and recreate the buffer
    rsa_tx["address"] = "Bis13AbAZwMeY1C5GuFuVuVKLSjr3RdKG63g4CEx6epwSbhpuDU3rj"
    buffer = str(
        (
            rsa_tx["timestamp"],
            rsa_tx["address"],
            rsa_tx["recipient"],
            rsa_tx["amount"],
            rsa_tx["operation"],
            rsa_tx["openfield"],
        )
    ).encode("utf-8")
    # Sign the tx
    signature = signer.sign_buffer_for_bis(buffer)
    print("ED25519 Signature", signature)

    public_key_hex = signer.to_dict()["public_key"]
    public_key_net = base64.b64encode(bytes.fromhex(public_key_hex)).decode("utf-8")
    print(public_key_hex, public_key_net)

    SignerFactory.verify_bis_signature(
        signature,
        public_key_net,
        buffer,
        "Bis13AbAZwMeY1C5GuFuVuVKLSjr3RdKG63g4CEx6epwSbhpuDU3rj",
    )
    print("No Error")



"""
Test bip32 btc like
"""

from coincurve import PrivateKey, PublicKey
from bip32utils import BIP32Key

# seed = os.urandom(32).hex()
seed = "8950b1cffcd85b6ad992feed179863da032b2674178127150db25f5208ea8e4a"
# TEST ONLY, DO NOT USE FOR REAL
print(seed)
pk = PrivateKey.from_hex(seed)
pkhex = pk.to_hex()
print("pkhex", pkhex)  # == SEED

key = BIP32Key.fromEntropy(seed.encode("utf-8"))
private_key = key.PrivateKey().hex()
extended_key = key.ExtendedKey()

public_key = key.PublicKey().hex()
address = key.Address()

print("private_key", private_key)
print("extended_key", extended_key)
print("public_key", public_key)
print("address", address)
identifier = key.Identifier().hex()
print("id", identifier)

priv = "d5ba601c19bfc54e109d20736453ba150f624ee03e2b7a5a3875342cd64d2efe"
seed = "e8f32e723decf4051aefac8e2c93c9c5b214313817cdb01a1494b917c8436b35"
pk2 = PrivateKey.from_hex(priv)
pub2 = pk2.public_key.format(compressed=False).hex()
pub2c = pk2.public_key.format(compressed=True).hex()  # This one is used
print("pub2", pub2)
print("pub2c", pub2c)



"""

"""

import base58
import hashlib
from polysign.signer import Signer, SignerType, SignerSubType
from typing import Union
from hashlib import sha256
from base64 import b64decode, b64encode
from coincurve import PrivateKey


class SignerCRW(Signer):

    __slots__ = ('_key', )

    _address_versions = {SignerSubType.MAINNET_REGULAR: b'\x01\x75\x07'}

    def __init__(self, private_key: Union[bytes, str]=b'', public_key: Union[bytes, str]=b'', address: str='',
                 compressed=False, subtype: SignerSubType=SignerSubType.MAINNET_REGULAR):
        super().__init__(private_key, public_key, address, compressed=compressed, subtype=subtype)
        self._key = None
        self._type = SignerType.ECDSA

    def from_private_key(self, private_key: Union[bytes, str], subtype: SignerSubType=SignerSubType.MAINNET_REGULAR):
        print('TODO')

    def from_full_info(self, private_key: Union[bytes, str], public_key: Union[bytes, str]=b'', address: str='',
                       subtype: SignerSubType = SignerSubType.MAINNET_REGULAR, verify: bool=True):
        print('TODO')

    def from_seed(self, seed: str='', subtype: SignerSubType = SignerSubType.MAINNET_REGULAR):
        print('crw from seed {}'.format(seed))
        if subtype != SignerSubType.MAINNET_REGULAR:
            self._subtype = subtype
        try:
            key = PrivateKey.from_hex(seed)
            public_key = key.public_key.format(compressed=False).hex()
            print("Public Key", public_key)
            self._key = key
            self._private_key = key.to_hex()  # == seed
            self._public_key = public_key
        except Exception as e:
            print("Exception {} reading RSA private key".format(e))
        print("identifier", self.identifier().hex())
        self._address = self.address()

    def identifier(self):
        """Returns double hash of pubkey"""
        # faafd1966c79c472360ef1cf8860169df6e7554a
        return hashlib.new('ripemd160', sha256(bytes.fromhex(self._public_key)).digest()).digest()

    def address(self):
        # 1PrWZ4CXSXWbg87XS9ShhwMV6TiSXtycT7
        vh160 = self.address_version_for_subtype(self._subtype) + self.identifier()  # raw content
        chk = sha256(sha256(vh160).digest()).digest()[:4]
        return base58.b58encode(vh160 + chk).decode('utf-8')

    @classmethod
    def public_key_to_address(cls, public_key: Union[bytes, str],
                              subtype: SignerSubType=SignerSubType.MAINNET_REGULAR) -> str:
        """Reconstruct an address from the public key"""
        if type(public_key) == str:
            identifier = hashlib.new('ripemd160', sha256(bytes.fromhex(public_key)).digest()).digest()
        else:
            identifier = hashlib.new('ripemd160', sha256(public_key).digest()).digest()
        vh160 = cls.address_version_for_subtype(subtype) + identifier  # raw content
        checksum = sha256(sha256(vh160).digest()).digest()[:4]
        return base58.b58encode(vh160 + checksum).decode('utf-8')

    @classmethod
    def verify_signature(cls, signature: Union[bytes, str], public_key: Union[bytes, str], buffer: bytes,
                         address: str='') -> None:
        """Verify signature from raw signature. Address may be used to determine the sig type"""
        raise ValueError("SignerCRW.verify_signature not impl.")

    @classmethod
    def verify_bis_signature(cls, signature: str, public_key: str, buffer: bytes, address: str = '') -> None:
        """Verify signature from bismuth tx network format (ecdsa sig and pubkey are b64 encoded)
        Returns None, but raises ValueError if needed."""
        public_key = b64decode(public_key).decode('utf-8')
        # print(public_key)

        """ TODO
        public_key_object = RSA.importKey(public_key_pem)
        signature_decoded = b64decode(signature)
        verifier = PKCS1_v1_5.new(public_key_object)
        sha_hash = SHA.new(buffer)
        if not verifier.verify(sha_hash, signature_decoded):
            raise ValueError(f"Invalid signature from {address}")
        """
        # Reconstruct address from pubkey to make sure it matches
        if address != cls.public_key_to_address(public_key):
            raise ValueError("Attempt to spend from a wrong address")

    def sign_buffer_raw(self, buffer: bytes) -> bytes:
        """Sign a buffer, sends a raw bytes array"""
        pass

    def sign_buffer_for_bis(self, buffer: bytes) -> str:
        """Sign a buffer, sends under the format expected by bismuth network format"""
        # sig is b64 encoded
        return b64encode(self.sign_buffer_raw(buffer))
